from pathlib import Path

import pythoncom
from PySide2.QtCore import QObject, Signal
from PySide2.QtWidgets import QTreeWidgetItem
from wmi import WMI

from hdlg.hdd import HDD
from hdlg.utils import size_unit, hdl_dump_live


class MainWorker(QObject):
    error = Signal(Exception)
    finished = Signal()
    progress = Signal(float)
    status_message = Signal(str)
    found_device = Signal(HDD)
    hdd_info = Signal(list)
    game = Signal(HDD, Path, str, str, str)

    def find_hdds(self) -> None:
        """
        Find Disk Drive devices using WMI on Windows.

        TODO: Add support for Linux and macOS.
        """
        try:
            self.status_message.emit("Scanning HDDs...")
            # noinspection PyUnresolvedReferences
            pythoncom.CoInitialize()  # important!
            c = WMI()
            disk_drives = c.Win32_DiskDrive()
            for disk_drive in sorted(disk_drives, key=lambda d: d.index):
                self.found_device.emit(HDD(
                    target=disk_drive.DeviceID,
                    model=disk_drive.Model
                ))
            self.status_message.emit(f"Found {len(disk_drives)} HDDs")
            self.finished.emit()
        except Exception as e:
            self.error.emit(e)

    def get_hdd_info(self, hdd: HDD) -> None:
        """Get HDD Usage Information like Total/Used/Available Disk Space and a list of Games."""
        try:
            self.status_message.emit(f"Loading HDD %s (%s)" % (hdd.target, hdd.model))
            disk_usage_percent = [
                (hdd.disk_map[1] / hdd.disk_map[0]) * 100,  # Used
                (hdd.disk_map[2] / hdd.disk_map[0]) * 100,  # Available
            ]
            space_tree = QTreeWidgetItem(["Disk Space"])
            space_tree.addChild(QTreeWidgetItem([
                "Total", f"{size_unit(hdd.disk_size)} ({hdd.disk_size})"
            ]))
            space_tree.addChild(QTreeWidgetItem([
                "Used", f"{size_unit(hdd.disk_map[1])} ({hdd.disk_map[1]}, {disk_usage_percent[0]:.2f}%)"
            ]))
            space_tree.addChild(QTreeWidgetItem([
                "Available", f"{size_unit(hdd.disk_map[2])} ({hdd.disk_map[2]}, {disk_usage_percent[1]:.2f}%)"
            ]))

            games = hdd.get_games_list()
            games_tree = QTreeWidgetItem(["Games", str(len(games))])
            for media_type, size, _, dma, game_id, name in games:
                games_tree.addChild(QTreeWidgetItem([
                    f"{media_type} {size_unit(size)} ({dma})",
                    f"{game_id} {name}"
                ]))

            self.hdd_info.emit([
                space_tree,
                games_tree
            ])
            self.status_message.emit(f"Loaded HDD %s (%s)" % (hdd.target, hdd.model))
            self.finished.emit()
        except Exception as e:
            self.error.emit(e)

    def install_game(self, hdd: HDD, iso: Path, media_type: str, disc_label: str, game_id: str):
        """Install a Game ISO to a PS2 HDD."""
        try:
            self.status_message.emit(f"Installing {iso.stem} ({game_id})")
            for line in hdl_dump_live(
                f"inject_{media_type.lower()}",
                hdd.hdl_target, iso.stem.title(), str(iso.absolute()), game_id
            ):
                progress = line.split(", ")
                if len(progress) == 3:
                    progress, remaining, speed = progress
                else:
                    progress, remaining, speed = progress[0], None, None
                self.status_message.emit(", ".join(x for x in [
                    f"{progress} Installed {iso.stem} ({game_id})", remaining, speed
                ] if x))
                progress = float(progress.split("%")[0])
                self.progress.emit(progress)
            self.status_message.emit("Installed %s (%s %s)..." % (disc_label, game_id, media_type))
            self.finished.emit()
        except Exception as e:
            self.error.emit(e)



import ctypes
import re
import shutil
import subprocess
import sys
from typing import Iterator

HDL_DUMP_BIN = shutil.which("hdl-dump") or shutil.which("hdl_dump")
NEIGHBORING_WHITESPACE = re.compile(r"[\s]{2,}")
CAMEL_TO_SNAKE_1 = re.compile(r"(.)([A-Z][a-z]+)")
CAMEL_TO_SNAKE_2 = re.compile(r"([a-z0-9])([A-Z])")
SIZE_UNITS = ["B", "KB", "MB", "GB", "TB", "PB"]


def is_admin() -> bool:
    """Check if running with Administrative rights."""
    try:
        return ctypes.windll.shell32.IsUserAnAdmin()
    except Exception as e:
        print(e)
        return False


def is_frozen() -> bool:
    """Check if running under frozen build (cx_freeze or PyInstaller)."""
    return getattr(sys, "frozen", False) and hasattr(sys, "_MEIPASS")


def require_admin() -> None:
    """If not running with Administrative rights, request it or close."""
    if not is_admin():
        ctypes.windll.shell32.ShellExecuteW(None, "runas", sys.executable, " ".join(sys.argv[is_frozen():]), None, 0)
        sys.exit(1)


def camel_to_snake(name: str) -> str:
    """
    Convert CamelCase to snake_case supporting a wide variety of scenarios.

    Examples:
        >>> camel_to_snake("CamelCaseName")
        camel_case_name
        >>> camel_to_snake("camel2_camel2_case")
        camel2_camel2_case
        >>> camel_to_snake("getHTTPResponseCode")
        get_http_response_code
        >>> camel_to_snake("HTTPResponseCodeXYZ")
        http_response_code_xyz
    """
    name = CAMEL_TO_SNAKE_1.sub("\1_\2", name)
    name = CAMEL_TO_SNAKE_2.sub("\1_\2", name)
    return name.lower()


def size_unit(size, base=1000):
    """
    Convert a file size (in bytes) to a Human Readable Prefixed Size Unit.

    Examples:
        >>> size_unit(131)
        131 B
        >>> size_unit(58812)
        58.81 KB
        >>> size_unit(68819826)
        68.81 MB
        >>> size_unit(39756861649)
        39.75 GB
        >>> size_unit(18754875155724)
        18.75 TB
    """
    i = 0
    while size >= base and i < len(SIZE_UNITS) - 1:
        size /= float(base)
        i += 1
    f = ("%.2f" % size).rstrip("0").rstrip(".")
    return "%s %s" % (f, SIZE_UNITS[i])


def hdl_dump(*args) -> list[str]:
    """Make a call to hdl-dump and return the string output."""
    res = subprocess.check_output([HDL_DUMP_BIN, *args])
    return res.decode().splitlines()


def hdl_dump_live(*args) -> Iterator[str]:
    """Make a call to hdl-dump and return every line as they are written to the std."""
    res = subprocess.Popen([HDL_DUMP_BIN, *args], stdout=subprocess.PIPE, bufsize=1, universal_newlines=True)
    while res.poll() is None:
        line = res.stdout.readline()
        if line:
            yield line.strip()



from __future__ import annotations

import re
import subprocess
import traceback
from pathlib import Path

from PySide2 import QtWidgets, QtGui, QtCore
from PySide2.QtWidgets import QMessageBox

from hdlg.hdd import HDD
from hdlg.ui import BaseWindow
from hdlg.ui.worker import MainWorker
from hdlg.utils import size_unit, HDL_DUMP_BIN


class Main(BaseWindow):
    def __init__(self):
        super().__init__(name=self.__class__.__name__)

        self.reset_state()
        self.window.setMinimumSize(1000, 400)
        self.window.hddInfoList.header().setSectionResizeMode(QtWidgets.QHeaderView.ResizeToContents)

        # menu bar actions
        self.window.actionExit.triggered.connect(self.window.close)
        self.window.actionAbout.triggered.connect(self.about)

        # button actions
        self.window.refreshIcon.clicked.connect(self.refresh_hdd_list)

        # store data here to keep in scope and out of garbage collection
        # necessary for the thread and worker class
        self.GC_KEEP = None

        self.refresh_hdd_list()

    def add_hdd_button(self, hdd: HDD) -> None:
        """Add an HDD button into the HDD list."""
        for child in self.window.deviceListDevices_2.children():
            if isinstance(child, QtWidgets.QPushButton):
                # Skip buttons with identical targets
                if child.objectName() == hdd.target:
                    return

        button = QtWidgets.QPushButton("\n".join([
            " ".join([
                hdd.hdl_target,
                f"({size_unit(hdd.disk_size)}" + [")", ", PS2)"][hdd.is_apa_partitioned]
            ]),
            hdd.model
        ]))
        button.setObjectName(hdd.target)
        button.setCursor(QtGui.QCursor(QtCore.Qt.PointingHandCursor))
        button.setEnabled(hdd.is_apa_partitioned)
        button.setCheckable(hdd.is_apa_partitioned)
        button.clicked.connect(lambda: self.load_hdd(hdd))
        button.toggled.connect(lambda: button.setChecked(True))

        device_list = self.window.deviceListDevices_2.layout()
        device_list.insertWidget(0 if hdd.is_apa_partitioned else device_list.count() - 1, button)

    def reset_state(self) -> None:
        """Reset the State of the UI and Application to the initial startup."""
        # Clear all HDD buttons from the HDD list
        for child in self.window.deviceListDevices_2.children():
            if isinstance(child, QtWidgets.QPushButton):
                # noinspection PyTypeChecker
                child.setParent(None)

        # Reset the HDD information panel
        self.window.hddInfoList.clear()
        self.window.hddInfoList.setEnabled(False)
        self.window.hddInfoList.addTopLevelItem(QtWidgets.QTreeWidgetItem([
            "\n" * 8 + " " * 90 +
            "Scanning for PS2 HDDs..."
        ]))

        # Reset the Installation Button
        try:
            # if you refresh after loading an HDD, you will end up with 2x receivers
            # so disconnect the oldest and leave the newest
            self.window.installButton.clicked.disconnect()
        except RuntimeError:
            # has no receiver, it's fine
            pass
        self.window.installButton.setEnabled(False)
        self.window.installButton.hide()

        # Reset the Progress Bar
        self.window.progressBar.hide()
        self.window.progressBar.setValue(0)

        # Enable the Refresh Button
        self.window.refreshIcon.setEnabled(True)

    def refresh_hdd_list(self) -> None:
        """Remove list of HDD devices and reload them."""
        self.reset_state()
        self.window.refreshIcon.setEnabled(False)

        thread = QtCore.QThread()
        worker = MainWorker()
        worker.moveToThread(thread)

        def on_finish():
            self.window.refreshIcon.setEnabled(True)
            self.window.hddInfoList.clear()
            self.window.hddInfoList.addTopLevelItem(QtWidgets.QTreeWidgetItem([
                "\n" * 8 + " " * 60 +
                "Ready to go? Just choose a PS2 HDD to get started!"
            ]))
            thread.quit()

        def on_error(e: Exception):
            msg = QMessageBox()
            msg.setIcon(QMessageBox.Critical)
            msg.setWindowTitle("Failed to list HDDs")
            msg.setText("An error occurred when listing HDDs:")
            msg.setDetailedText(traceback.format_exc())
            msg.setInformativeText(str(e))
            msg.exec_()

        worker.finished.connect(on_finish)
        worker.error.connect(on_error)

        worker.status_message.connect(self.window.statusbar.showMessage)
        worker.found_device.connect(self.add_hdd_button)

        thread.started.connect(worker.find_hdds)
        thread.start()

        self.GC_KEEP = (thread, worker)

    def load_hdd(self, hdd: HDD):
        """Load HDD device, get HDD object, get device information."""
        # prevent refreshing of HDDs, loading of an HDD, or installation
        self.window.refreshIcon.setEnabled(False)
        self.window.deviceListDevices_2.setEnabled(False)
        self.window.installButton.hide()
        if self.window.installButton.isEnabled():
            self.window.installButton.clicked.disconnect()

        self.window.hddInfoList.clear()
        self.window.hddInfoList.addTopLevelItem(QtWidgets.QTreeWidgetItem([
            "\n" * 8 + " " * 100 +
            "Loading PS2 HDD..."
        ]))

        thread = QtCore.QThread()
        worker = MainWorker()
        worker.moveToThread(thread)

        def on_finish():
            self.window.deviceListDevices_2.setEnabled(True)
            self.window.refreshIcon.setEnabled(True)
            self.window.installButton.setEnabled(True)
            self.window.installButton.show()
            self.window.hddInfoList.setEnabled(True)
            self.window.installButton.clicked.connect(lambda: self.install_game(hdd))
            thread.quit()

        def on_error(e: Exception):
            msg = QMessageBox()
            msg.setIcon(QMessageBox.Critical)
            msg.setWindowTitle("Failed to load HDD")
            msg.setText("An error occurred when loading the HDD:")
            msg.setDetailedText(traceback.format_exc())

            error = None
            if isinstance(e, subprocess.CalledProcessError):
                error = e.output.decode()
                if error and "unrecognized command" in error:
                    error = "Unrecognized command, hdl-dump version may be too old."
                elif e.returncode == 107:
                    error = "APA partition is broken"

            if error:
                msg.setInformativeText(error)
            else:
                msg.setInformativeText(str(e))

            msg.exec_()

        def set_hdd_info(trees: list[QtWidgets.QTreeWidgetItem]):
            self.window.hddInfoList.clear()
            for tree in trees:
                self.window.hddInfoList.addTopLevelItem(tree)
            self.window.hddInfoList.expandToDepth(0)

        worker.finished.connect(on_finish)
        worker.error.connect(on_error)

        worker.status_message.connect(self.window.statusbar.showMessage)
        worker.hdd_info.connect(set_hdd_info)

        thread.started.connect(lambda: worker.get_hdd_info(hdd))
        thread.start()

        self.GC_KEEP = (thread, worker)

    def install_game(self, hdd: HDD):
        filenames = QtWidgets.QFileDialog.getOpenFileNames(
            self.window,
            "Select one or more PS2 Games to Install",
            filter="All supported formats (*.ISO *.ZSO *.GI *.CUE *.IML *.NRG);;"
                   "ISO images (*.ISO);;"
                   "ZSO Compressed ISO images (*.ZSO);;"
                   "CDRWIN cuesheets (*.CUE);;"
                   "Nero Burning Rom images (*.NRG);;"
                   "RecordNow! Global Image (*.GI);;"
                   "Sony CD/DVD Intermediate files (*.IML);;"
                   "All files (*.*)",
            # dir=str(cfg.user_cfg.last_opened_directory or "")
        )
        if not filenames[0]:
            self.log.debug("Cancelled Installation as no PS2 Disc Image (ISO) was provided.")
            return
        filenames = [Path(x) for x in filenames[0]]

        def _install(index: int = 0):
            if index > len(filenames) - 1:
                return
            iso_path = filenames[index]
            # TODO: Verify info cdvd_info
            try:
                cdvd_info = subprocess.check_output([
                    HDL_DUMP_BIN,
                    "cdvd_info2",
                    str(iso_path)
                ], stderr=subprocess.PIPE)

                disc_info = re.match(r'^(?:dual-layer )?([^ ]*) +(\d+)KB +"([^"]*)" +"([^"]+)"', cdvd_info.decode())
                if not disc_info:
                    QMessageBox.information(
                        self.window,
                        "Unable to Identify Game Data",
                        f"Skipping \"{iso_path}\" as game information like Media Type and ID could not be identified.\n"
                        f"\n" +
                        cdvd_info.decode()
                    )
                    _install(index + 1)
                    return
                media_type, game_size, disc_label, game_id = disc_info.groups()

                self.window.deviceListDevices_2.setEnabled(False)
                self.window.refreshIcon.setEnabled(False)
                self.window.installButton.setEnabled(False)
                self.window.progressBar.show()
                self.window.progressBar.setValue(0)

                thread = QtCore.QThread()
                worker = MainWorker()
                worker.moveToThread(thread)

                def on_progress(n: float):
                    self.window.progressBar.setValue(n)

                def on_finish():
                    self.window.deviceListDevices_2.setEnabled(True)
                    self.window.refreshIcon.setEnabled(True)
                    self.window.installButton.setEnabled(True)
                    # update disk usage and games list
                    self.load_hdd(hdd)
                    thread.quit()
                    _install(index + 1)

                def on_error(e: Exception):
                    msg = QMessageBox()
                    msg.setIcon(QMessageBox.Critical)
                    msg.setWindowTitle("Failed to install Game")
                    msg.setText("An error occurred when installing a Game to an HDD:")
                    msg.setDetailedText(traceback.format_exc())
                    msg.setInformativeText(str(e))
                    msg.exec_()

                worker.progress.connect(on_progress)
                worker.finished.connect(on_finish)
                worker.error.connect(on_error)

                worker.status_message.connect(self.window.statusbar.showMessage)

                # TODO: Is there a better way to do this?
                worker.game.connect(worker.install_game)
                thread.started.connect(lambda: worker.game.emit(hdd, iso_path, media_type, disc_label, game_id))
                thread.start()

                self.GC_KEEP = (thread, worker)
            except subprocess.CalledProcessError as e:
                QMessageBox.information(
                    self.window,
                    "Cannot Install Game",
                    f"Skipping \"{iso_path}\" due to an error: {e.stderr.decode().strip()} [{e.returncode}]"
                )
                _install(index + 1)

        _install()



import logging
import platform
import struct
import sys
from datetime import datetime

from PySide2 import QtCore
from PySide2.QtUiTools import QUiLoader
from PySide2.QtWidgets import QMessageBox

from hdlg.config import Directories
from hdlg.utils import camel_to_snake, is_frozen


class BaseWindow:
    def __init__(self, name: str, flag=QtCore.Qt.Window) -> None:
        name = camel_to_snake(name)

        loader = QUiLoader()
        loader.setWorkingDirectory(QtCore.QDir(str(Directories.root)))
        ui_file = QtCore.QFile(str(Directories.root / "ui" / f"{name}.ui"))
        ui_file.open(QtCore.QFile.ReadOnly)

        self.window = loader.load(ui_file)
        self.window.setWindowFlags(flag)
        ui_file.close()

        self.log = logging.getLogger(name)

    def show(self) -> None:
        self.window.show()

    def about(self) -> None:
        # TODO: Get information from pyproject.toml
        QMessageBox.about(
            self.window,
            "About HDLG",
            ("HDLG v0.2.1 [%s]" % (
                ",".join(map(str, filter(None, [
                    sys.platform,
                    "%dbit" % (8 * struct.calcsize("P")),
                    platform.python_version(),
                    "frozen" if is_frozen() else None
                ])))
            )) +
            f"<p>Copyright (C) {datetime.now().year} PHOENiX</p>" +
            ("<p>{0}<br/><a href='{1}' style='color:white'>{1}</a></p>".format(
                "Modern GUI for hdl-dump.",
                "https://github.com/rlaphoenix/hdlg"
            ))
        )



"""
hdlg - Modern GUI for hdl-dump.
Copyright (C) 2021-2022 rlaphoenix

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <https://www.gnu.org/licenses/>.
"""

from pathlib import Path

from appdirs import AppDirs


class Directories:
    app_dirs = AppDirs("hdlg", False)
    root = Path(__file__).resolve().parent  # root of package/src



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['hdlg', 'hdlg.ui']

package_data = \
{'': ['*'], 'hdlg.ui': ['icons/*']}

install_requires = \
['PySide2>=5.15.2.1,<6.0.0.0',
 'WMI>=1.5.1,<2.0.0',
 'appdirs>=1.4.4,<2.0.0',
 'pywin32==304',
 'toml>=0.10.2,<0.11.0']

entry_points = \
{'console_scripts': ['hdlg = hdlg.hdlg:main']}

setup_kwargs = {
    'name': 'hdlg',
    'version': '0.2.1',
    'description': 'Modern GUI for hdl-dump.',
    'long_description': '![Banner](https://rawcdn.githack.com/rlaphoenix/hdlg/50bab8126da83a63e83bf6a5ce3d4d1f737ced2b/banner.png)\n\n[![Build status](https://github.com/rlaphoenix/hdlg/actions/workflows/ci.yml/badge.svg)](https://github.com/rlaphoenix/hdlg/actions/workflows/ci.yml)\n[![PyPI version](https://img.shields.io/pypi/v/hdlg)](https://pypi.python.org/pypi/hdlg)\n[![Python versions](https://img.shields.io/pypi/pyversions/hdlg)](https://pypi.python.org/pypi/hdlg)\n<a href="https://github.com/rlaphoenix/hdlg/blob/master/LICENSE">\n  <img align="right" src="https://img.shields.io/badge/license-GPLv3-blue" alt="License (GPLv3)"/>\n</a>\n\nHDLG is a modern GUI for hdl-dump with Batch installation capabilities.\n\n**hdl-dump**: <https://github.com/ps2homebrew/hdl-dump>  \n**wLaunchELF**: <https://github.com/ps2homebrew/wLaunchELF>\n\n![Preview](https://user-images.githubusercontent.com/17136956/198822365-f244dcf6-3050-45f2-83dd-c32c4d36f976.png)  \n*Preview as of October 2022.*\n\n## Installation\n\n    pip install --user hdlg\n\nTo run hdlg, type `hdlg` into any terminal, command prompt, app launcher, or the start menu.\n\nIf you wish to manually install from the source, take a look at [Building](#building-source-and-wheel-distributions).\n\n## To-do\n\n- [x] Craft initial GUI with Qt.\n- [x] Push to PyPI and add relevant Badges.\n- [x] Add PyInstaller make file.\n- [x] Add local PS2 HDD connection option.\n- [x] List installed games of selected HDD.\n- [x] Show HDD information like Disk Size, Space Used, and such.\n- [x] Add ability to install a new game to selected HDD.\n- [x] Add batch installation support by selecting more than one file.\n- [ ] Create a file based settings system.\n- [ ] Add per-install settings like startup, flags, and DMA mode.\n- [ ] Add ability to format an HDD for use with a PS2 with `pfsshell`.\n- [ ] Add ability to rename the Game Name of installed games.\n- [ ] Add ability to extract an installed game from the PS2 HDD.\n- [ ] Add ability to view an installed game\'s sector table.\n- [ ] Add ability to set a custom icon to an installed game.\n- [ ] Add remote PS2 HDD (samba) connection option.\n- [ ] Add Inno Setup script.\n- [ ] Add Linux support.\n- [ ] Add macOS support.\n\n## Building\n\nThis project requires [Poetry], so feel free to take advantage and use it for its various conveniences like\nbuilding sdist/wheel packages, creating and managing dependencies, virtual environments, and more.\n\nNote:\n\n- Source Code may have changes that may be old, not yet tested or stable, or may have regressions.\n- Only run or install from Source Code if you have a good reason. Examples would be to test for regressions, test\n  changes (either your own or other contributors), or to research the code (agreeing to the [LICENSE](LICENSE)).\n- [Poetry] is required as it\'s used as the [PEP 517] build system, virtual environment manager, dependency manager,\n  and more.\n\n  [Poetry]: <https://python-poetry.org/docs/#installation>\n  [PEP 517]: <https://www.python.org/dev/peps/pep-0517>\n\n### Install from Source Code\n\n    git clone https://github.com/rlaphoenix/hdlg.git\n    cd hdlg\n    pip install --user .\n\n### Building source and wheel distributions\n\n    poetry build\n\nYou can specify `-f` to build `sdist` or `wheel` only. Built files can be found in the `/dist` directory.\n\n### Packing with PyInstaller\n\n    poetry run python pyinstaller.py\n\nThe build is now available at `./dist`.\n',
    'author': 'PHOENiX',
    'author_email': 'rlaphoenix@pm.me',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/rlaphoenix/hdlg',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<3.11',
}


setup(**setup_kwargs)



"""
hdlg - Modern GUI for hdl-dump.
Copyright (C) 2021-2022 rlaphoenix

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <https://www.gnu.org/licenses/>.
"""

import os
import sys

from PySide2 import QtCore
from PySide2.QtWidgets import QApplication

from hdlg.config import Directories
from hdlg.ui.main import Main
from hdlg.utils import require_admin


def main():
    require_admin()

    os.environ["QT_AUTO_SCREEN_SCALE_FACTOR"] = "1"
    QtCore.QCoreApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling)

    app = QApplication(sys.argv)
    app.setStyle("fusion")
    app.setStyleSheet((Directories.root / "ui" / "app.qss").read_text("utf8"))

    window = Main()
    window.show()

    sys.exit(app.exec_())


if __name__ == "__main__":
    main()



"""
hdlg - Modern GUI for hdl-dump.
Copyright (C) 2021-2022 rlaphoenix

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <https://www.gnu.org/licenses/>.
"""

from __future__ import annotations

import struct
from ctypes.wintypes import HANDLE
from pathlib import Path
from typing import Union

import win32con
import win32file
import winioctlcon

from hdlg.utils import NEIGHBORING_WHITESPACE, hdl_dump


class HDD:
    def __init__(self, target: Union[str, Path], model: str):
        self.handle = win32file.INVALID_HANDLE_VALUE
        self.target = target
        self.hdl_target = target.upper().replace(r"\\.\PHYSICALDRIVE", "hdd") + ":"
        self.model = model

        self._geometry = None
        self._disk_size = None
        self._disk_map = None
        self._is_apa_partitioned = None
        self._apa_checksum = None

        self.open(target)

    def __enter__(self):
        return self

    def __exit__(self, *_, **__):
        self.dispose()

    def dispose(self):
        if self.handle != win32file.INVALID_HANDLE_VALUE:
            win32file.CloseHandle(self.handle)

    def open(self, device: str, extended=False) -> HANDLE:
        if not device.startswith("\\\\.\\"):
            device = r"\\.\%s" % device  # unc target
        self.handle = win32file.CreateFile(
            # https://docs.microsoft.com/en-us/windows/win32/api/fileapi/nf-fileapi-createfilea
            device,  # target
            win32con.MAXIMUM_ALLOWED,  # 0x0080 | 0x0020,  # desired access
            win32con.FILE_SHARE_READ | win32con.FILE_SHARE_WRITE,  # share mode
            None,  # security attributes
            win32con.OPEN_EXISTING,  # creation disposition
            win32con.FILE_ATTRIBUTE_NORMAL,  # flags and attributes
            None  # template file
        )
        if self.handle == win32file.INVALID_HANDLE_VALUE:
            raise RuntimeError("Failed to obtain device handle...")
        if extended:
            # TODO: Sometimes works, sometimes doesn't. Only got it working (once) on a CD-ROM drive with a DVD.
            print(win32file.DeviceIoControl(self.handle, winioctlcon.FSCTL_ALLOW_EXTENDED_DASD_IO, None, None))
        return self.handle

    def seek(self, to: int, whence: int = win32file.FILE_BEGIN) -> int:
        pos = win32file.SetFilePointer(self.handle, to, whence)
        if pos != to:
            raise IOError(f"Seek was not precise...")
        return pos

    def read(self, size: int) -> bytes:
        if size % 512 != 0:
            raise ValueError("Size must be a multiple of 512 for some reason, Ask Windows.")
        res, data = win32file.ReadFile(self.handle, size, None)
        if res != 0:
            raise IOError(f"An error occurred: {res} {data}")
        if len(data) < size:
            raise IOError(f"Read {size - len(data)} less bytes than requested...")
        return data

    @property
    def geometry(self) -> tuple[int, ...]:
        """
        Retrieves information about the physical disk's geometry.
        https://docs.microsoft.com/en-us/windows/win32/api/winioctl/ns-winioctl-disk_geometry_ex

        Returns a tuple of:
            Cylinders-Lo
            Cylinders-Hi
            Media Type
            Tracks Per Cylinder
            Sectors Per Track
            Bytes Per Sector
            Disk Size
            Extra Data
        """
        if self._geometry is not None:
            return self._geometry

        if self.handle == win32file.INVALID_HANDLE_VALUE:
            raise ValueError("Handle is invalid, have you opened a handle yet?")

        self._geometry = struct.unpack("8L", win32file.DeviceIoControl(
            self.handle,  # handle
            winioctlcon.IOCTL_DISK_GET_DRIVE_GEOMETRY_EX,  # ioctl api
            None,  # in buffer
            32  # out buffer
        ))
        return self._geometry

    @property
    def disk_size(self) -> int:
        """Get full Disk Size (in bytes)."""
        if self._disk_size is not None:
            return self._disk_size

        cyl_lo, cyl_hi, _, tpc, spt, bps, _, _ = self.geometry

        self._disk_size = (cyl_lo + cyl_hi) * tpc * spt * bps
        return self._disk_size

    @property
    def disk_map(self) -> tuple[int, ...]:
        """Get Total Slice Size, Used Space, and Available Space (in bytes)."""
        if self._disk_map is not None:
            return self._disk_map

        disk_map = hdl_dump("map", self.hdl_target)[-1]
        total, used, available = [
            int(x.split(": ")[1][:-2]) * 1000 * 1000
            for x in disk_map.split(", ")
        ]

        self._disk_map = (total, used, available)

        return self._disk_map

    @property
    def is_apa_partitioned(self) -> bool:
        """Check if HDD is a PS2 APA-formatted device."""
        if self._is_apa_partitioned is not None:
            return self._is_apa_partitioned

        old_pos = self.seek(0, whence=win32file.FILE_CURRENT)

        try:
            if self.seek(0) != 0:
                raise RuntimeError("Unable to seek to start of HDD, cannot check if APA Partitioned...")
            header = self.read(1024)
            checksum = header[0:4]
            magic = header[4:8]

            new_checksum = struct.pack(
                "<Q",
                sum(struct.unpack("<I", header[n:n+4])[0] for n in range(4, 1024, 4))
            )[:4]

            self._is_apa_partitioned = magic == b"APA\0" and checksum == new_checksum
        finally:
            self.seek(old_pos)

        return self._is_apa_partitioned

    @property
    def apa_checksum(self):
        if not self.is_apa_partitioned:
            return None

        if self._apa_checksum is not None:
            return self._apa_checksum

        old_pos = self.seek(0, whence=win32file.FILE_CURRENT)

        try:
            if self.seek(0) != 0:
                raise RuntimeError("Unable to seek to start of HDD, cannot get APA checksum...")
            header = self.read(512)
            self._apa_checksum = header[0:4]
        finally:
            self.seek(old_pos)

        return self._apa_checksum

    def get_games_list(self) -> list[tuple[str, int, int, str, str, str]]:
        """
        Get a list of games installed on the HDD (if any).

        Returns a tuple of:
            MediaType
            Size (in Bytes)
            Flags
            DMA
            GameID
            GameName
        """
        # TODO: Add handler for when there's no games installed
        games = hdl_dump("hdl_toc", self.hdl_target)[1:-1]
        # [!] will show as the Game Name for any game that was improperly installed
        games = [
            (NEIGHBORING_WHITESPACE.sub(" ", game).split(" ", maxsplit=5) + ["[!]"])[:6]
            for game in games
        ]
        games = [
            (media_type, int(size.replace("KB", "")) * 1000, int(flags), dma, game_id, game_name)
            for media_type, size, flags, dma, game_id, game_name in games
        ]
        games.sort(key=lambda x: x[-1])
        return games
